/*
    Sweeper ... 
	This file is a part of sweeper, a program for identifying
	and estimating regression models, selecting from many possible 
	predictors without overfitting.

    Copyright (C) 1999,2002  Robert A. Stine

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License in the file COPYING for more details.
*/
// $Id: fileOps.c,v 1.13 2000/11/20 22:37:50 bob Exp $

/*
    26 Jun 00 ... Add ability to generate a file with signal.
    14 Jul 99 ... Add write of selected columns and interactions back to file.
    11 Jun 99 ... Fix interaction terms: dont want to cross leading terms(its Y)!
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "fileOps.h"
#include "random.h"

#ifndef TRUE
#define FALSE 0
#define TRUE 1
#endif

#define ExitOnNull(ptr,str,result) if(NULL==(ptr)){fprintf(stderr,"\nFILEOPS: %s\n",str);return (result);}


/****************************************************************************/

int WriteRandomFile (char *fileName, int nLines, int nCols)
{
  int i;
  FILE *outFile;

  // Check status of generator
  if (! HaveValidRandomState() )
  { fprintf(stderr, "Need to initialize random generator first.\n");
    return 0;
  }
  
  // Open file for writing, deleting prior contents if any
  outFile = fopen(fileName, "w+");
  if (outFile == NULL)
    return -1;
  
  while (nLines--)
  {
    // write random weight
    fprintf (outFile, "%f ", UniformRand() );
    // now add data cols
    for (i=0; i<nCols; ++i)
    	fprintf(outFile, "%6.3f ", NormalRand() );
    fprintf(outFile,"\n");
  }
  fclose(outFile);
  return (nCols);
}
			

int WriteRegrFileWithSignal (char *fileName, int nLines, int nCols, 
			       int predCol, float beta)
{
  float x1, y, xPred;
  int i;
  FILE *outFile;
  
  // Check status of generator
  if (! HaveValidRandomState() )
  { fprintf(stderr, "Need to initialize random generator first.\n");
    return 0;
  }

  // Open file for writing, deleting prior contents if any
  outFile = fopen(fileName, "w+");
  if (outFile == NULL)
    return -1;
  
  while (nLines--)
  {
    // generate the x1 and chosen pred and associated probability
    xPred = NormalRand();
    x1    = NormalRand();
    y     = 1000.0 + beta/4.0 * x1 + beta * xPred + NormalRand();

    // write observation weight
    fprintf (outFile, "%f ", UniformRand()); 
    // write response and first predictor used in interaction
    fprintf(outFile, "%6.3f %6.3f ", y, x1);
    // write rest of preds, with saved predictor value inserted
    for (i=2; i<predCol; ++i)
      fprintf(outFile, "%6.3f ", NormalRand() );
    // write predictor and one that is correlated with it
    fprintf(outFile, "%6.3f %6.3f ", xPred, xPred+0.2*NormalRand());
    for (i=predCol+2; i<nCols; ++i)
      fprintf(outFile, "%6.3f ", NormalRand() );
    fprintf(outFile,"\n");
  }
  fclose(outFile);
  return (nCols);
}


int WriteBooleanFileWithSignal (char *fileName, int nLines, int nCols, 
			       int predCol, float effectSize)
{
  double alpha = 3.0, beta = 2.0;
  float x1, y, xPred, prob, weight;
  int i;
  FILE *outFile;
  
  // Check status of generator
  if (! HaveValidRandomState() )
  { fprintf(stderr, "Need to initialize random generator first.\n");
    return 0;
  }

  // Open file for writing, deleting prior contents if any
  outFile = fopen(fileName, "w+");
  if (outFile == NULL)
    return -1;
  
  while (nLines--)
  {
    // generate the x1 and chosen pred and associated probability
    xPred = beta * GammaRand(alpha);    
    x1    = NormalRand();
    prob = 1.0/ (1.0 + exp(- ((effectSize * (xPred - 1.75 * alpha * beta))
			      + ( 0.0 * xPred * x1))));
    // now see if random y is large enough
    if (UniformRand() < prob)
    { y = 1.0;
      weight = 1.0;
    } else
    { y = 0.0;
      weight = 1.0;
    }
    // write observation weight
    fprintf (outFile, "%f ", weight ); 
    // write response and first predictor used in interaction
    fprintf(outFile, "%6.3f %6.3f ", y, x1);
    // write rest of preds, with saved predictor value inserted
    for (i=2; i<predCol; ++i)
      fprintf(outFile, "%6.3f ", NormalRand() );
    // write predictor and one that is correlated with it
    fprintf(outFile, "%6.3f %6.3f ", xPred, xPred+0.2*NormalRand());
    for (i=predCol+2; i<nCols; ++i)
      fprintf(outFile, "%6.3f ", NormalRand() );
    fprintf(outFile,"\n");
  }
  fclose(outFile);
  return (nCols);
}


/****************************************************************************/


int CountFields (char *fileName, int lineNumber)
{
  int counter, ch;
  int readingField;
  FILE *fp;
  
  // Open file as read only
  fp = fopen(fileName, "r");
  if (fp == NULL)
  {	fprintf(stderr, "FileOPS: file %s not found\n", fileName);
	return -1;
  };	
  // Skip the desired number of lines
  while (--lineNumber > 0)
    while ((ch = getc(fp)) != '\n')
      if (ch == EOF) return (-2);			
  // Read line, counting the number of space-delimited fields.
  counter = 0;
  readingField = FALSE;
  while ((ch = getc(fp)) != '\n')
  {	// printf("Read %c reading field %d -> ", ch, readingField);
    if (isspace(ch))
    { 	if (readingField == TRUE)
    {	++counter; 
	readingField = FALSE;
    }
    }
    else if (readingField == FALSE)
      readingField = TRUE;		
    // printf("%d\n", readingField);
  }
  if (readingField) ++ counter;
  // Close file and return total number of columns
  fclose(fp);
  return (counter);
}


int CountLines (char *fileName)
{	int ch;
	long line; 
    FILE *fp;
	
	// Open file as read only
	fp = fopen(fileName, "r");
	if (fp == NULL)
	{	return -1;
	};	
	// count the number of lines as the number of \n found
	line = 0;
	while (EOF != (ch = getc(fp)))
		if (ch == '\n') ++ line;
	// Close file and return number of lines
	fclose(fp);
	return (line);
}



/****************************************************************************/

Vector MeansFromTextFile (char *fileName, int nColumns)
{	double buffer, sumWeights;
	float weight;
	int i, line, cont;
	Vector means;
	FILE *fp;
	
	// Open file as read only
	fp = fopen(fileName, "r");
	if (fp == NULL)
	{	return NULL;
	};
	// Check space for means and initialize
	means = NewVector(nColumns, "MeansFromTextFile");
	if (NULL == means)
	{	fprintf (stderr, "FileOps: Allocation of means failed.\n");
		fclose(fp);
		return NULL;
	} 
	InitializeVector(means, 0.0);
    // Loop over lines of the file
    sumWeights = line = 0;
    cont = fscanf(fp, "%f", &weight);
    while (cont != EOF)
    {	for (i=0; i < nColumns; ++i)
    	{	fscanf(fp, "%lf", &buffer);
    		IncrementVectorElement(means, i, weight * buffer);
    	};
    	++line;
    	sumWeights += weight;
    	cont = fscanf(fp, "%f", &weight);
    };
    // Normalize by count
    for (i=0; i<nColumns; ++i)
     	DivideVectorElement(means, i, sumWeights);
    fclose(fp);
    // printf ("Read %d lines.\n", line);
    return(means);
}


/****************************************************************************/


int ConvertToBinaryFile (char *asciiFileName, int nColumns, char *binFileName)
{	int i, line, cont, outCont, nChars;
	FILE *input, *output;
	char *binData;
	float *buffer;
	
	// Exit if not enough buffer space.  Make room for floating case weight.
	buffer = new float[nColumns];
	if (buffer == NULL)
	{	return -77;
	};
	
	// Open input text file as read only, output binary file for writing.
	input = fopen(asciiFileName, "r");
	if (input == NULL)
	{	delete(buffer);
		return -1;		
	};
	output = fopen(binFileName, "w+");
	if (output == NULL)
	{	delete(buffer);
		return -2;
	};
    // Read input line, writing collection of values when accumulated.
    //   Assumes that the input file has equal num of cols per line, starting with 
    //   float sampling weight which is read first.
    line = 0;
    nChars = nColumns * (int) sizeof(float);
    while ((cont = fscanf(input, "%f", &buffer[0])) != EOF)
    {	for (i=1; i < nColumns; ++i)
    	 	cont = fscanf(input, "%f", &buffer[i]);
    	binData = (char *) buffer;
    	for (i=0; i < nChars; ++i)
    		outCont = fputc (binData[i], output);
    	++line;
    }
    // delete buffer, close files, and return
    delete(buffer);
    fclose(input);
    fclose(output);
    return line;
}

/********************* char buffer *************************************/

static FILE *theFile;

#define OpenTheFile(name)  if((theFile=fopen(name,"r"))==NULL) return 0;
#define CloseTheFile()     fclose(theFile);

int LoadCharBuffer(char *pBuffer, long nChars);
// Loads input array with nChar characters from the file, using fgetc


int LoadCharBuffer(char *pBuffer, long nChars)
{
	*pBuffer = (char) fgetc(theFile);
	if (*pBuffer == EOF)
	  	return EOF;
	else
	{	while (--nChars)
			*(++pBuffer) = (char) fgetc(theFile);
	}
	return 1;
}

/****************************************************************************/

/*
	Number of columns for output covariances as function of number of input file columns
	If interactions, then adjust for Y leading term (linear + quad + pairs)
*/
#define OUTPUT_COLS(interact, n) (interact) ? (n) + ((n)-1) + (((n)-1)*((n)-2))/2 : (n) 

/********************** float buffer ***********************************
 
  Allocate/Loads input array with floats from the file, returning sampling weight.
  Read program extracts with the proper adjustment for interactions.

*/

static float *theFloatBuffer = NULL;
static long lengthOfFloatBuffer = 0;


int    AllocFloatBuffer(long length);
double LoadFloatBuffer(long nFloats);
void   FloatBufferIndices(long index, long *ii, long *jj);
float  FloatBuffer(long index);				// allows for interaction index
int    DeleteFloatBuffer();

int AllocFloatBuffer (long length)
{
    if (NULL != theFloatBuffer)
    {	fprintf (stderr, "\n\nAllocating non-null float buffer that was not nil.\n");
    	delete(theFloatBuffer);
    }
    theFloatBuffer = new float[length];
    if (NULL == theFloatBuffer)
    {	fprintf(stderr, "Buffer allocation failed in mean.\n");
    	return 0;
    }
	else 
	{	lengthOfFloatBuffer = length;
		return 1;
	}
}
	
double LoadFloatBuffer(long nFloats)
{	float weight;
	unsigned long nRead;

	nRead = fread((void *)&weight, sizeof(float), 1, theFile);
	if(1 == nRead)
	{	fread((void *)theFloatBuffer, sizeof(float), (unsigned long) nFloats, theFile);
		return weight;
	} else
	return -1.0;
}


void FloatBufferIndices(long index, long *ii, long *jj)
{
	long iiOffset;
		
	index -= lengthOfFloatBuffer;		// get past linear
	*ii = 1;                                // data_1 is first predictor column
	iiOffset = 0;
	*jj = lengthOfFloatBuffer-1;            // len - 1 is number of predictors
	while (*jj)
	{	if (index < iiOffset + *jj)	// found in this block
		{	*jj = *ii + index - iiOffset;	// jj >= ii
			return ;
		}
		iiOffset += *jj;
		++*ii;
		--*jj;
	}
}


float FloatBuffer(long index)
{
	if (index < lengthOfFloatBuffer)
		return theFloatBuffer[index];
	else
	{	long ii, jj;
		FloatBufferIndices(index, &ii, &jj);
		return (theFloatBuffer[ii]*theFloatBuffer[jj]);
	}
}


int DeleteFloatBuffer()
{
	delete(theFloatBuffer);
	theFloatBuffer = NULL;
	return 1;
}



Vector MeansFromBinaryFile (char *fileName, long nFileColumns, int interact, long *nLines, double *nWeightedObs)
{
  long i, j, line, nOutCols, index;
  double weight;
  double sumWeights;
  Vector means;
  
  nOutCols = OUTPUT_COLS(interact,nFileColumns);
  // Check space for means, initialize buffer.
  means = NewVector(nOutCols, "MeansFromBinaryFile");
  ExitOnNull(means, "Allocation of means failed.", NULL);
  AllocFloatBuffer(nFileColumns);
  // Open file as read only
  OpenTheFile(fileName);
  // Initialize totals
  InitializeVector(means, 0.0);
  // Loop over lines of the file
  sumWeights = line = 0;
  while ((weight = LoadFloatBuffer(nFileColumns)) > 0.0)
  {
    for (i=0; i < nFileColumns; ++i)
      IncrementVectorElement(means, i, weight * theFloatBuffer[i]);
    if (interact)  // calc mean of raw cross product
    {	index = nFileColumns;
	for (i=1; i < nFileColumns; ++i)  // start at one to skip Y
	  for (j=i; j<nFileColumns; ++j)  // raw access - no index to interaction area
	    IncrementVectorElement(means, index++,
				   weight * theFloatBuffer[i] * theFloatBuffer[j]);
    	}
    	++line;
    	sumWeights += weight;
    };
    // Normalize by count
    for (i=0; i<nOutCols; ++i)
     	DivideVectorElement(means, i, sumWeights);
    // Close file and release buffer
    CloseTheFile();
    DeleteFloatBuffer();
    // fprintf(stderr, "Computed means from %d lines.\n", line);
    *nLines = line;
    *nWeightedObs = sumWeights;
    return(means);
}
    

/****************************************************************************/

	
Vector SSFromBinaryFile (char *fileName, long nFileColumns, Vector means, int interact)
{
  double dev, sumWeights;
  long i, j, line,  nOutCols, destIndex;
  double weight;
  Vector ss;
  
  nOutCols = OUTPUT_COLS(interact,nFileColumns);
  // Check space for means, initialize buffer.
  if (nOutCols != VectorLength(means))
  {	fprintf(stderr, "Length of means vector does not match file size.\n");
    	return NULL;
  }
  ss = NewVector(nOutCols, "SSFromBinaryFile");
  ExitOnNull(ss, "Allocation of ss failed.", NULL);
  AllocFloatBuffer(nFileColumns);
  // Open file as read only
  OpenTheFile(fileName);
  InitializeVector(ss, 0.0);
  // Loop over lines of the file
  sumWeights = line = 0;
  while ((weight = LoadFloatBuffer(nFileColumns)) >  0.0)
  {
    for (i=0; i < nFileColumns; ++i)
    {	dev = theFloatBuffer[i] - VectorElement(means, i);
	IncrementVectorElement(ss, i, weight * (dev * dev));
    }
    if (interact)
    {	destIndex = nFileColumns;
	for (i=1; i < nFileColumns; ++i)	// start at one to skip Y
	{
	  for (j=i; j<nFileColumns; ++j)	// direct indexing here
	  { dev = (theFloatBuffer[i]*theFloatBuffer[j]) -
	      VectorElement(means, destIndex);
	    IncrementVectorElement(ss, destIndex, weight * (dev * dev));
	    ++destIndex;
	  }
	}
    }
    ++line;
    sumWeights += weight;
  }
  // Close file and release buffer
  CloseTheFile();
  DeleteFloatBuffer();
  // fprintf(stderr, "Computed SS from %d lines.\n", line);
  return(ss);
}
    
	
Vector CPFromBinaryFile (char *fileName, long nFileColumns, Vector means, long crossColumn, int interact)
{
  double dev, weight, sumWeights;
  long i, j, line, nOutCols, destIndex;
  Vector cp;
  
  nOutCols = OUTPUT_COLS(interact, nFileColumns);
  // Check space for means, initialize buffer.
  if (VectorLength(means) != nOutCols || crossColumn > nOutCols || crossColumn  < 0)
  {	fprintf(stderr, "Length or index error in CPFromBinaryFile.\n");
    	return NULL;
  }
  cp = NewVector(nOutCols, "CPFromBinaryFile");
  ExitOnNull(cp, "Allocation of cp failed.",NULL);
  AllocFloatBuffer(nFileColumns);
  // Open file as read only
  OpenTheFile(fileName);
  InitializeVector(cp, 0.0);
  // Loop over lines of the file
  sumWeights = line = 0;
  if (interact)
    while ((weight = (double) LoadFloatBuffer(nFileColumns)) > 0.0)
    {	dev = FloatBuffer(crossColumn) - VectorElement(means, crossColumn);
	for (i=0; i < nFileColumns; ++i)
	  IncrementVectorElement(cp, i, weight * (dev * theFloatBuffer[i]));
	destIndex = nFileColumns;
	for (i=1; i < nFileColumns; ++i)	// start at one to skip Y
	{
	  for (j=i; j<nFileColumns; ++j)
	  { IncrementVectorElement(cp, destIndex,
				   weight * dev*theFloatBuffer[i]*theFloatBuffer[j]);
	    ++destIndex;
	  }
	}
	++line;	
	sumWeights += weight;
    }
  else
    while ((weight = (double) LoadFloatBuffer(nFileColumns)) > 0.0)
    {	dev = theFloatBuffer[crossColumn] - VectorElement(means, crossColumn);
	for (i=0; i < nFileColumns; ++i)
	  IncrementVectorElement(cp, i, weight * (dev * theFloatBuffer[i]));
	++line;	
	sumWeights += weight;
    }
  // Normalize by count-1
  /*
    --sumWeights;
    for (i=0; i<nColumns; ++i)
    cov[i] /= (double) sumWeights;
  */
  // Close file and release buffer
  CloseTheFile();
  DeleteFloatBuffer();
  // fprintf(stderr, "Computed CP from %d lines.\n", line);
  return(cp);
}

    	
int ComputeCPsFromBinaryFile (char *fileName, long nFileColumns, Vector means,
			      long *pCols, int nCols, Vector vecArray[], int interact)
{	
  double dev, weight, incr, sumWeights;
  long i, j, k, line, col, nOutCols, destIndex;
  Vector vec;
	
  // Figure output space needs
  nOutCols = OUTPUT_COLS(interact, nFileColumns);

  // Initialize buffer
  AllocFloatBuffer(nFileColumns);
  if (theFloatBuffer == NULL) return -77;
	
  // Check input space for resulting vector elements, initialize to zero
  for (i=0; i<nCols; ++i)
  {
    vecArray[i] = NewVector(nOutCols, "ComputeCPsFromBinaryFile");
    ExitOnNull(vecArray[i], "Failed allocating covariance vector", -7);
    InitializeVector(vecArray[i], 0.0);
  }
  // Open file as read only
  OpenTheFile(fileName);
  // Loop over lines of the file
  sumWeights = line = 0;
  if (interact)
    while ((weight = (double) LoadFloatBuffer(nFileColumns)) > 0.0)
    {
      for (k=0; k<nCols; ++k)
      {	col = pCols[k];
	vec = vecArray[k];
	dev = FloatBuffer(col) - VectorElement(means, col);
	for (i=0; i < nFileColumns; ++i)	// include Y here
	  IncrementVectorElement(vec, i, weight * (dev * theFloatBuffer[i]));
	destIndex = nFileColumns;
	for (i=1; i < nFileColumns; ++i)	// start at one to skip Y
	{
	  for (j=i; j<nFileColumns; ++j)
	  { incr = weight * (dev * theFloatBuffer[i] * theFloatBuffer[j]);
	    if (incr > 1.0E25 || incr < -1.0E25)
	    { printf ("Working on dest %ld for col %ld\n", destIndex, col);
	      printf ("Incr %f out of range for i=%ld, j=%ld, weight = %f, dev = %f\n",
		      incr,i,j,weight,dev);
	    }
	    IncrementVectorElement(vec, destIndex,
				   weight * dev*theFloatBuffer[i]*theFloatBuffer[j]);
	    ++destIndex;
	  }
	}
      }
      ++line;	
      sumWeights += weight;
    }
  else
    while ((weight = (double) LoadFloatBuffer(nFileColumns)) > 0.0)
    {
      for (k=0; k<nCols; ++k)
      {	col = pCols[k];
	vec = vecArray[k];
	// center the value for column i
	dev = theFloatBuffer[col] - VectorElement(means, col);
	// calc cross-prods with other cols
	for (i=0; i < nFileColumns; ++i)
	  IncrementVectorElement(vec, i, weight * (dev * theFloatBuffer[i]));
      }
      ++line;
      sumWeights += weight;
    }
  // Close file and release buffer
  CloseTheFile();
  DeleteFloatBuffer();
  return(nCols);
}

// --------------------------------------------------------------------------------
// Computes the linear combination from the file

double SSOfLinearCombinationFromBinaryFile (char *fileName, long nFileColumns,
					    Vector coefs, long *coefIndex,
					    double offset, Vector weights)
{
  double lpSS, lp, fileWeight;
  long i, line, c1, c2, nPredictors;
  
  AllocFloatBuffer(nFileColumns);
  // Indexing based on number of possible predictors
  nPredictors = nFileColumns -1 ;
  // Open file as read only
  OpenTheFile(fileName);
  // Loop over lines of the file
  lpSS = 0.0;
  line = 0;
  while ((fileWeight = (double) LoadFloatBuffer(nFileColumns)) > 0.0)
  { lp = offset;
    for (i=0; i < coefs->len; ++i)
    {	IndexToFileColumns(coefIndex[i], nPredictors, &c1, &c2);
	if(c2)
	  lp += VectorElement(coefs, i) * theFloatBuffer[c1] * theFloatBuffer[c2];
	else
	  lp += VectorElement(coefs, i) * theFloatBuffer[c1];
    }
    if (weights)
    {	lpSS += VectorElement(weights, line) * fileWeight * (lp * lp);
	++line;
    }
    else
      lpSS += fileWeight * lp * lp;
  }
  // Close file and release buffer
  CloseTheFile();
  DeleteFloatBuffer();
  // fprintf(stderr, "Computed CP from %d lines.\n", line);
  return(lpSS);
}

Vector LinearCombinationFromBinaryFile (char *fileName, long nFileColumns,
					Vector coefs, long *coefIndex,
					double offset, long nFileLines, Vector weights)
{
  Vector linComb;
  double lc, fileWeight;
  long i, line, c1, c2, nPredictors;
  
  AllocFloatBuffer(nFileColumns);
  linComb =NewVector(nFileLines,"LinCombBinaryFile");
  if (NULL == linComb) 
  { fprintf(stderr, "FileOps: Could not alloc vector with %ld values.\n", nFileLines);
    return (NULL);
  }
  // Indexing based on number of possible predictors
  nPredictors = nFileColumns -1 ;
  // Open file as read only
  OpenTheFile(fileName);
  // Loop over lines of the file
  line = 0;
  while ((fileWeight = (double) LoadFloatBuffer(nFileColumns)) > 0.0)
  { lc = offset;
    for (i=0; i < coefs->len; ++i)
    { IndexToFileColumns(coefIndex[i], nPredictors, &c1, &c2);
      if(c2)
      	lc += VectorElement(coefs, i) * theFloatBuffer[c1] * theFloatBuffer[c2];
      else
      	lc += VectorElement(coefs, i) * theFloatBuffer[c1];
    }
    if (weights) 
    { lc *= fileWeight * VectorElement(weights, line);
      ++line;
    }
    else
      lc *= fileWeight;
    AssignVectorElement(linComb, line, lc);
    ++line;
  }
  // Close file and release buffer
  CloseTheFile();
  DeleteFloatBuffer();
  // fprintf(stderr, "Computed CP from %d lines.\n", line);
  return(linComb);
}



Matrix CPMatrixFromBinaryFile (char *fileName, long nFileColumns, Vector means, long *pCols, long nCols, Vector weights)
{
  Matrix cpMatrix;
  double dev, incr, fileWeight, weight;
  long i, j, line, nPredictors;
  
  AllocFloatBuffer(nFileColumns);
  cpMatrix = NewMatrix(nCols, nCols, "CPMatrixFromBinaryFile");
  if (NULL == cpMatrix) 
  { fprintf(stderr, "FileOps: Could not allocate matrix for CP.\n");
    return (NULL);
  }
  // Indexing based on number of possible predictors
  nPredictors = nFileColumns -1 ;
  // Open file as read only
  OpenTheFile(fileName);
  // Loop over lines of the file
  line = 0;
  while ((fileWeight = (double) LoadFloatBuffer(nFileColumns)) > 0.0)
  {
    for (i=0; i < nCols; ++i)
    { dev = FloatBuffer(pCols[i]) - VectorElement(means, pCols[i]);
      if (weights) 
	weight = fileWeight * VectorElement(weights, line);
      else weight = fileWeight;
      IncrementMatrixElement(cpMatrix, i, i, weight * dev * dev);
      for (j=1; j < nCols; ++j)
      {	incr = weight * dev * FloatBuffer(pCols[j]);
	IncrementMatrixElement(cpMatrix, i, j, incr);
	IncrementMatrixElement(cpMatrix, j, i, incr);
      }
    }
    ++line;
  }
  // Close file and release buffer
  CloseTheFile();
  DeleteFloatBuffer();
  // fprintf(stderr, "Computed CP from %d lines.\n", line);
  return(cpMatrix);
}


			
// --------------------------------------------------------------------------------
// Second index will be zero if a simple predictor

void IndexToFileColumns (long index, long nPreds, long *x1, long *x2)
{
  *x2 = 0;
  if (index <= nPreds)				// simple predictor
    *x1 = index;
  else
  { index -= nPreds;
    while (index + *x2 > nPreds)	// so that index + *x2 <= p when break
    { index -= nPreds - *x2;
      ++ *x2;
    }
    *x1 = index + *x2;
    ++ *x2;
  }
}


long WriteColumnsToFile (char *inFile, long nFileColumns, char *outFileName, 
			 long *outputCols, long nVars)
{
  long i, j, x1, x2, line;
  double weight;
  FILE *outFile;
  
  AllocFloatBuffer(nFileColumns);
  OpenTheFile(inFile);	
  // Open file for writing, deleting prior contents if any
  outFile = fopen(outFileName, "w+");
  if (outFile == NULL)
    return -1;
  // Loop over lines of the file
  line = 0;
  while ((weight = LoadFloatBuffer(nFileColumns)) > 0.0)
  { fprintf (outFile, "%5.2f ", weight); 
    for (i=0; i < nVars; ++i)
    { j = outputCols[i];
      IndexToFileColumns(j, nFileColumns-1, &x1, &x2);  // -1 to not count Y
      if (x2)		// interaction
	fprintf(outFile, "%f ", theFloatBuffer[x1]*theFloatBuffer[x2]);
      else 		// output simple deviation
	fprintf(outFile, "%f ", theFloatBuffer[j]);
    }
    fprintf(outFile, "\n");
    ++line;
  }
  // Close file and release buffer
  CloseTheFile();
  DeleteFloatBuffer();
  fclose(outFile);
  // fprintf(stderr, "Computed SS from %d lines.\n", line);
  return(line);
}
