/*
    Sweeper ... 
	This file is a part of sweeper, a program for identifying
	and estimating regression models, selecting from many possible 
	predictors without overfitting.

    Copyright (C) 1999,2002  Robert A. Stine

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License in the file COPYING for more details.
*/
// $Id: fileOps.h,v 1.6 2000/08/07 01:59:17 bob Exp $

/*
	13 Sep 99 ... Remove the growing amount of binary data file object.
	 7 Sep 99 ... Add binary file, weights, linear combination, vector/matrix.
	28 Jul 99 ... Add LP of linear combination function.
	14 Jul 99 ... WriteColumnsToFile, IndexToFileColumns added.
	10 Jun 99 ... Start to handle pairwise interactions in file computations.
	30 Oct 98 ... Convert to return vectors.
*/

#ifndef _FILEOPS_
#define _FILEOPS_

#include "vector.h"
#include "matrix.h"

// ---------------------------------------------------------------------------------
// 
//   Basic ascii file operations
//
//    nCols argument includes the response
// 		
// --------------------------------------------------------------------------------- 
   
int WriteRandomFile (char *fileName, int nLines, int nCols);
/*
	Writes a file with uniform random values. Returns column count if
	all is OK.  Includes a uniform sampling weight integer as the first item,
	so you get a total of nCols + 1 columns written to the output.

	Be sure that the random number generator has been initialized
*/

int WriteRegrFileWithSignal (char *fileName, int nLines, int nCols, 
			     int predCol, float beta);
/*
	Writes file as 'WriteRandomFile', but adds a signal of the form
	     y = 1000 + beta/4 x1 + beta xPred + N(0,1)
	into file.  
*/


int WriteBooleanFileWithSignal (char *fileName, int nLines, int nCols, 
			       int predCol, float effectSize);  
/*
	Writes file as 'WriteRandomFile', but adds a signal component to the
	file.  The response column (Y) which comes after the weights is 0/1
	with pr(y=1) = logistic(a=0, b=effectSize). 
*/

int CountFields (char *fileName, int lineNumber);
/*	
	Counts number of space delimited column fields found in the open file,
	using values found in indicated line. 
*/

int CountLines (char *fileName);
/*	
	Counts the number of lines found in the text file. 
*/

Vector MeansFromTextFile (char *fileName, int nColumns);
/*	
	Computes the column means of the indicated file and return these
	as a vector with indicated number of columns. Treats first column as weights.
*/

int ConvertToBinaryFile (char *asciiFileName, int nColumns, char *binFileName);
/*
	Converts the indicated input file into a packed binary file for 
	later applications, using only single precision (4-bytes). All columns
	treated as floats and written to output. (ie, no special first columns).
*/

// ---------------------------------------------------------------------------------
// 
//   Binary data file operations
//
// ---------------------------------------------------------------------------------

long WriteColumnsToFile (char *inFile, long nFileColumns,
			 char *outFile, long *cols, long nVars);
/*
	Write selected from the indicated *binary* input file to the output ascii file,
	allowing for interactions between file columns.  The input file is assumed to
	have a weight as the first column, and this is automatically copied over.
	Column 0 is the response column, the first column after the weights.
	
	Returns the number of lines actually written.
*/

void IndexToFileColumns (long index, long nPreds, long *x1, long *x2);
/*
	Converts the input index into two integers.  The second integer is 0 if
	the index is a simple column.  The pair x1,x2 otherwise identify the
	columns of an interaction term.
*/

Vector MeansFromBinaryFile (char *fileName, long nFileColumns,
			    int interact, long *nLines, double *nWeightedObs);
Vector SSFromBinaryFile (char *fileName, long nFileColumns,
			 Vector means, int interact);
Vector CPFromBinaryFile (char *fileName, long nFileColumns,
			 Vector means, long column, int interact);
/*	
	Computes the column characteristic of the indicated BINARY file, placing
	in the input vector.  Input vector must have room for the indicated number of
	columns.  Sums of sqares and cross-products require means to be pre-computed. 
	Cross-products are of the form Cov(j,i), i=1,...,nColumns.
	
	The interact flag signals whether to compute all of the pairwise
	interactions of the indicated columns.
	
	Find the weighted number of observations via means calculation.
*/

int ComputeCPsFromBinaryFile (char *fileName, long nFileColumns,
			      Vector means, long *pCols, int nCols,
			      Vector vecArray[], int interact);
/*	
	Computes nCols cross-products at once from the input file. 
	Vector elements must be allocated and of sufficient size for covariances.
*/

Vector LinearCombinationFromBinaryFile (char *fileName, long nFileColumns,
					Vector coefs, long *coefIndex,
					double offset, long nFileLines,
					Vector weights);

double SSOfLinearCombinationFromBinaryFile (char *fileName, long nFileColumns,
					    Vector coefs, long *coefIndex,
					    double offset, Vector weights);
/*
	Computes the linear combination of the columns defined by the coefs
	vector with the indexing implied by the cols pointer, and accumulates
	the sum of the squared values over the observations
	in the indicated binary file.  The value of the offset is added to
	each linear combination prior
	to squaring.

	Automatically computes the interaction terms as needed based on
	coefIndex values.
	
	If weights vector is null, then the function uses file weights.
	If the vector is "full"	then uses these weights *times* those in the file.
	
	Examples:
	To get the residual SS,
	  Set coefs[0] to -1 and coefs[1..k] to regression slopes,
	  with offset as -intercept.

	To get the fitted values,
	  Set coefs[0]=0.0 and rest to regr slopes, with offset as intercept.
*/

Matrix CPMatrixFromBinaryFile (char *fileName, long nFileColumns,
			       Vector means, long *pCols, long nCols, Vector weights);
/*
	Computes the cross-product matrix from the indicated file, automatically
	adding interactions where needed in the calculations.
*/

#endif
