/*
    Sweeper ... 
	This file is a part of sweeper, a program for identifying
	and estimating regression models, selecting from many possible 
	predictors without overfitting.

    Copyright (C) 1999,2002  Robert A. Stine

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License in the file COPYING for more details.
*/
// $Id: parser.c,v 1.12 2002/03/15 15:14:52 bob Exp $

#include <stdio.h>
#include <string>
#include <iostream>

#include "parser.h"
#include "sweepMatrix.h"
#include "binaryDataFile.h"
#include "fileOps.h"
#include "memory.h"

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif


#define BUFFER 100

static clock_t startTime, stopTime;


//-----------------------------------------------------------------------

void PrintDuration(clock_t start, clock_t finish)
{
  printf ("\nRequired %5.1f sec. \n",
	  ((double) (finish - start)) / CLOCKS_PER_SEC);
  printf ("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");
}


// ---------------------------------------------------------------------
//    This routine prompts for a user reply; extract from gBuffer

#define gBufferSize 60
static char gBuffer[gBufferSize];

int PromptLine (char *prompt);

//-----------

int PromptLine (char *prompt)
{	
  int c=0, i;
  
  printf("%s", prompt);
  for (i=0; i<gBufferSize-1 && (c = getchar()) != EOF && c !='\n'; ++i)
    gBuffer[i] = c;
  if (EOF == c)		// Exit
    return (0);
  else
  { gBuffer[i++] = '\0';
    return i;
  }
}


// -----------------------------------------------------------------------
#define NUM_TO_ADD 200

void Parser (char *path, int useInteraction, int useWts)
{
  BinaryDataFile bdf=NULL, validationBDF = NULL;
  SweepMatrix sm=NULL, wSM=NULL;
  Matrix covMat=NULL;
  Vector wts;
  double sumWts;
  char buff[gBufferSize],
    fileName[gBufferSize], 
    *prompt = "b/oir/aAt/pd/ws/fq> ",
    command;
  int arg;
  long xx[NUM_TO_ADD];


  // This version is hard wired to use the input command path rather
  // than an alternative at this level.
  
  // Parser first requires a localized path to use
  // PromptLine("\nPlease input path name for data files (e.g. test, bank): ");
  // strcpy (path, basePath); strcat (path, gBuffer); strcat(path,"/");
   
  command = 'x';
  while (command != 'q')
  { printf("%s", prompt);
    std::cin >> command;
    startTime = clock();
    switch (command)
    {      
    case 'b':
      printf ("Building binary data files in %s \n", path);
      bdf = MakeBinaryDataFile (path, "est", useWts, useInteraction);
      if (bdf) 
      { PrintSummaryBDF (bdf);
	SaveBinaryDataFile (bdf);}
      else printf ("Could not create initial BDF.\n");
      validationBDF = MakeBinaryDataFile (path, "val",  useWts, useInteraction);
      if (validationBDF) 
      { PrintSummaryBDF (validationBDF); SaveBinaryDataFile (validationBDF);}
      else printf ("Could not create validation BDF.\n");
      break;
    case 'o' :
      printf("Opening BDF on path %s...\n", path);
      strcpy(fileName, path); strcat(fileName, "est");
      bdf = OpenBinaryDataFile (fileName);
      if (bdf)
	PrintSummaryBDF (bdf);
      else printf (" Cannot open bin data file with prefix %s\n", fileName);
      strcpy(fileName, path); strcat(fileName,"val");
      validationBDF = OpenBinaryDataFile (fileName);
      if (validationBDF)
	PrintSummaryBDF (validationBDF);
      else
	printf ("Could not open validation BDF with prefix %s\n", fileName);
      break;
    case 'i':
      printf("Initializing sweep matrix and attaching to BDF.\n");
      if (sm)
      { printf (" ---> Freeing previous sweep matrix\n");
	delete(sm);
      }
      if (bdf)
      { sm = NewSweepMatrix (path, bdf, validationBDF);
	if (sm)
	{ ReadVariableNamesFromFile(sm);
	  PrintSweepMatrix(sm);
	}
	else printf ("*** Could not build sweep matrix. ***\n");
      }
      else printf(" ---> Need to open a BDF first\n");
      break;
    case 'u':
      printf ("Replacing prior sweep matrix with reweighted matrix\n");
      wts = BinomialVector(sm, &sumWts);
      wSM = ReweightedSweepMatrix(sm, wts);
      //  delete the old bdfs, use those from updated sm
      DeleteSweepMatrix (sm); 
      DeleteBinaryDataFile(bdf); DeleteBinaryDataFile(validationBDF);
      DeleteVector(wts);
      sm = wSM;
      bdf = sm->bdf; validationBDF = sm->validationBDF;
      break;
    case 'a':
      std::cin >> arg;
      printf ("Adding %d variables to model...\n", arg);
      if (sm && arg > 0)
      { ExpandStepwiseModel(sm, arg);
	WriteCurrentFit(sm);
	WriteCoefficientsToFile(sm);
      }
      else printf (" ---> Use i to initialize sweep matrix first");
      break;
    case 'A':
      std::cin >> arg;
      if (arg > NUM_TO_ADD)
	printf ("Trying to add too many into model\n");
      else
	printf ("Adding %d variables to model...\n", arg);
      for (int i=0; i<arg; ++i)
	std::cin >> xx[i];
      if (sm)
      { AddPredictorsToModel(sm, xx, arg, TRUE); // force these terms into model
	WriteCurrentFit(sm);
	WriteCoefficientsToFile(sm);
      }
      else printf (" ---> Use i to initialize sweep matrix first");
      break;

    case 'L':        // Force items into covariance cache    3/15/02
      std::cin >> arg;
      if (arg > NUM_TO_ADD)
	printf ("Trying to add too many into model\n");
      else
	printf ("Adding %d variables to model...\n", arg);
      for (int i=0; i<arg; ++i)
	std::cin >> xx[i];
      if (sm)
      { int nAdded = UpdateCache(sm, xx, arg, TRUE); // force these terms into model
        if (0 > nAdded) fprintf(stderr, "Error loading covariance queue.\n");
      }
      else printf (" ---> Use i to initialize sweep matrix first");
      break;

    case 't':
      std::cin >> arg;
      printf ("Trimming %d variables from model...", arg);
      if (sm)
	TrimStepwiseModel(sm, arg);
      else
	printf (" ---> Use i to initialize sweep matrix first\n");
      break;
      
    case 'p':
      printf ("Printing summary of sweep matrix\n");
      if (sm)
	PrintSweepMatrix(sm);
      else printf(" ---> Use i to initialize a sweep matrix first");
      break;

    case 'c':
      if (sm)
      { printf ("Computing covariance matrix for fitted slopes\n");
	if ((int) NumberOfPredictors(sm))
	{ std::cin >> arg;
	  if (arg)
	    printf ("\n Matrix from extended sweep form (full sweep)");
	  else
	    printf ("\n Matrix from X'X inversion");
	  covMat = RawCovMatOfEstimates (sm, arg);
	  MultiplyMatrixByConstant(covMat, 1000.0);
	  PrintMatrix (covMat, "1000 s^2 (X'X)^-1");
	  DeleteMatrix (covMat);
	} else printf("---> fit a model first\n");
      }
      else printf(" ---> Use i to initialize a sweep matrix first");  
      break;
      
    case 'w':
      printf ("Writing model and summary to log file\n");
      if (sm)
	WriteCurrentFit(sm);
      else printf(" ---> Use i to initialize the sweep matrix first");
      break;
    case 's':
      printf ("Writing model summary to log file\n");
      if (sm)
	WriteCurrentFitSummary(sm);
      else printf(" ---> Use i to initialize sweep matrix first");
      break;
      
    case 'd':
      printf ("Saving model data to regr.dat file\n");
      if (sm)
	WriteRegressionDataToFile(sm);
      else printf(" ---> Build a sweep matrix first");
      break;
    case 'f':
      printf ("Saving sweep matrix to file ");
      strcpy(fileName, "sweep-model");
      if (sm)
      { std::cin >> buff;
	strcpy(fileName, path); strcat(fileName, buff);
	printf("%s\n", fileName);
	SaveSweepMatrixToFile (sm, fileName);
      }
      else printf ("--> make a sweep matrix first\n");
      break;
    case 'r':
      if (sm)
      { printf ("Release (delete) current sweep matrix\n"); 
	DeleteSweepMatrix(sm);
      }
      std::cin >> buff;
      strcpy(fileName, path); strcat(fileName, buff);
      printf ("Restoring sweep matrix from file %s \n", fileName);
      sm = RestoreSweepMatrixFromFile (fileName);
      break;
    case 'm':
      printf ("Printing memory use...\n");
      PrintMemoryUse();
      break;
    case 'q':
      printf ("\nQuitting sweeper\n");
      break;
    default:
      printf ("\nDid not recognize the input command; commands are\n");
      printf ("  b    Build underlying binary data file \n");
      printf ("  o    Open binary data file for analysis \n");
      printf ("  i    Initialize sweep matrix from binary data file\n");
      printf ("  a #  Add indicated number of variables to sweep matrix \n");
      printf ("  A p x1 x2 ... xp  Add indicated predictors to sweep\n");
      printf ("  L p x1 x2 ... xp  Load indicated predictors into cache\n");
      printf ("  t #  Trim variables from sweep matrix \n");
      printf ("  u    Update weights of sweep matrix \n");
      printf ("  s    Summary of the model \n");
      printf ("  w    Write model description to file \n");
      printf ("  f s  Write sweep matrix to file\n");
      printf ("  r s  Restore sweep matrix from file\n");
      printf ("  p    Print summary of model \n");
      printf ("  m    Print current memory use \n");
      printf ("  q    Quit \n");
      break;
    }
    printf("  \n  ");
    stopTime = clock();
    CheckMemoryUse(TRUE);
    PrintDuration (startTime, stopTime);
  }
}



