/*
    Sweeper ... 
	This file is a part of sweeper, a program for identifying
	and estimating regression models, selecting from many possible 
	predictors without overfitting.

    Copyright (C) 1999,2002  Robert A. Stine

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License in the file COPYING for more details.
*/
// $Id: predFilter.c,v 1.4 2001/01/18 01:38:09 bob Exp $

/*
   1 Oct 00 ... Created from binary data file to read an ascii pipe

   Reads from an input stream on standard io, with following format
   separated on different lines.  The first k+2 lines define the model,
   and the remaining n+1 give the data.  
   
          q = number of variables in model
	  0 b0 intercept
	  col1 b1
	  col2 b2
	  ...
	  colq bq
	  n p  = number of cases, number of predictors
	  data for case 1 as y,x1,x2,...,xp
	  data for case 2
	  ...
	  data for case n

   It is able to build the interaction terms from large indices.

   Writes the first line (all to standard output) as
        validationSS minPred maxPred
   with the calibration information arranged as
        lower limit for prediction, #0, #1
   The sequence of lower limits gives a partition, with range formed
   as
         lower <= y^ < upper

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector.h"

// Use this switch to handle files that include weights (as in test cases)

#define WEIGHTS


/*********************** Global constants   *************************************/ 

static long p, q; // number of possible and actual predictors in model
static float gWeight=1.0;
static float *gBuffer;

/*********************** Internal functions *************************************/ 

void PredictorsUsedInColumn(long index, long *p1, long *p2);
void LoadBuffer (void);
double BufferElement (long index);

void PredictorsUsedInColumn(long index, long *p1, long *p2)
{
  *p2 = 0;
  if (index <= p)			// simple predictor
    *p1 = index;
  else
  {
    index -= p;
    while (index + *p2 > p)	        // so that index + *x2 <= p when break
    {	index -= p - *p2;
	++ *p2;
    }
    *p1 = index + *p2;
    ++ *p2;
  }
}

void LoadBuffer (void)
{
  long i;
  float *pBuff;

  pBuff = gBuffer;
  
#ifdef WEIGHTS
  scanf ("%f ", &gWeight);
#endif
  for (i=0; i<p; ++i) // response goes in position 0
    scanf("%f ", pBuff++);
  scanf("%f\n", pBuff);
}


double BufferElement (long index)
{
  long p1, p2;
  PredictorsUsedInColumn(index, &p1, &p2);
  if (p2)
    return (gBuffer[p1] * gBuffer[p2]);
  else
    return (gBuffer[p1]);
}


/********************************************************************************/    

#define LOWER_LIMIT  -1.0
#define UPPER_LIMIT   2.0
#define TABLE_SIZE   3000

int main (void)
{
  Vector coefs;
  double prediction, error, sse;
  double limit, delta, maxPred, minPred;
  double limits[TABLE_SIZE];
  long zeroCount[TABLE_SIZE], oneCount[TABLE_SIZE];
  double dIndex;
  
  long n, i, line, index;
  long *pColumns;

  // read model order and allocate space
  scanf ("%ld\n", &q);
#ifdef WEIGHTS
  fprintf(stderr, "PredFilter: Reading WEIGHTED model with %ld coefs ... ", q);
#else
  fprintf(stderr, "PredFilter: Reading model with %ld coefs ... ", q);
#endif
  coefs = NewVector (q+1, "main");
  pColumns = new long[q+1];
  // read (index coefs) pairs
  for (i=0; i<=q; ++i)
    scanf ("%ld %lf\n", &pColumns[i], &coefs->ptr[i]);
  fprintf(stderr, "done\n");
  // read p and then n on next line
  scanf ("%ld\n%ld\n", &p, &n);
  // allocate buffer for input
  fprintf(stderr, "PredFilter: Reading %ld obs for %ld predictors\n", n, p);
  gBuffer = new float[p+1];

  // build table and initialize counts
  delta = (UPPER_LIMIT - LOWER_LIMIT)/(double) TABLE_SIZE;
  for (i=0, limit=LOWER_LIMIT; i<TABLE_SIZE; ++i)
  { limits[i] = limit;
    zeroCount[i] = 0;
    oneCount[i] = 0;
    limit += delta;
  }
    
  // main calculation loop
  sse = 0.0;
  minPred=1.0;
  maxPred=0.0;
  for (line=0; line<n; ++line)
    for (i = 0; i < q; ++i)
    { LoadBuffer();
      prediction = VectorElement(coefs,0);
      for (i=1; i < coefs->len; ++i)
	prediction += VectorElement(coefs, i) * BufferElement(pColumns[i]);
      error = BufferElement(0) - prediction;
      // check for extremes
      if (prediction < minPred) minPred = prediction;
      else if (prediction > maxPred) maxPred = prediction;
      // insert into table
      dIndex = (prediction - LOWER_LIMIT)/delta;
      index = (long) dIndex;
      if (index < 0) index = 0;
      else if (index >= TABLE_SIZE) index = TABLE_SIZE-1;
      if (0 == BufferElement(0))
	++ zeroCount[index];
      else
	++ oneCount[index];
#ifdef WEIGHTS
      sse += gWeight * error * error;
#else
      sse += error * error;
#endif
    }

  // write the results to standard out
  printf ("%10.3f %10.3f %10.3f\n", sse, minPred, maxPred);
  for (i=0; i<TABLE_SIZE; ++i)
    printf ("%6.4f %ld %ld\n", limits[i], zeroCount[i], oneCount[i]);

  // write a note to the screen in case piped
#ifdef WEIGHTS
  fprintf(stderr, "PredFilter: Weighted Total SS = %10.3f\n", sse);
#else
  fprintf(stderr, "PredFilter: Total SS = %10.3f\n", sse);
#endif
  return 1;
}

