/*
    Sweeper ... 
	This file is a part of sweeper, a program for identifying
	and estimating regression models, selecting from many possible 
	predictors without overfitting.

    Copyright (C) 1999,2002  Robert A. Stine

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License in the file COPYING for more details.
*/
// $Id: predict.c,v 1.2 2002/06/01 20:57:56 bob Exp $

/*
  9 May 02 ... Modified from predFilter to produce a table with yHat and y pairs.
  1 Oct 00 ... Created from binary data file to read an ascii pipe.

   Reads from an input stream on standard io, with following format
   separated on different lines.  The first k+2 lines define the model,
   and the remaining n+1 give the data.  
   
          q = number of variables in model
	  0 b0 intercept
	  col1 b1
	  col2 b2
	  ...
	  colq bq
	  n p  = number of cases, number of predictors
	  data for case 1 as y,x1,x2,...,xp
	  data for case 2
	  ...
	  data for case n

   It is able to build the interaction terms from large indices.

   Writes a line (to standard error) as
        validationSS minPred maxPred
   followed by writing  the calibration information (to standard output) arranged as
               yHat y
   pairs on each line of output.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "vector.h"

// Use this switch to handle files that include weights (as in test cases)

#define WEIGHTS


/*********************** Global constants   *************************************/ 
static int p;
static float gWeight=1.0;
static float *gBuffer;

/*********************** Internal functions *************************************/ 

void PredictorsUsedInColumn(long index, long *p1, long *p2);
void LoadBuffer (void);
double BufferElement (long index);

void PredictorsUsedInColumn(long index, long *p1, long *p2)
{
  *p2 = 0;
  if (index <= p)			// simple predictor
    *p1 = index;
  else
  {
    index -= p;
    while (index + *p2 > p)	        // so that index + *x2 <= p when break
    {	index -= p - *p2;
	++ *p2;
    }
    *p1 = index + *p2;
    ++ *p2;
  }
}

void LoadBuffer (void)
{
  long i;
  float *pBuff;

  pBuff = gBuffer;
  
#ifdef WEIGHTS
  scanf ("%f ", &gWeight);
#endif
  for (i=0; i<p; ++i) // response goes in position 0
    scanf("%f ", pBuff++);
  scanf("%f\n", pBuff);
}


double BufferElement (long index)
{
  long p1, p2;
  PredictorsUsedInColumn(index, &p1, &p2);
  if (p2)
    return (gBuffer[p1] * gBuffer[p2]);
  else
    return (gBuffer[p1]);
}


/********************************************************************************/    

int main (void)
{
  Vector coefs;
  double prediction, sse, max, min;
  int    q,n, *pColumns;

  // read model order and allocate space
  scanf ("%d\n", &q);
#ifdef WEIGHTS
  fprintf(stderr, "predict: Reading WEIGHTED model with %d coefs ... ", q);
#else
  fprintf(stderr, "predict: Reading model with %ld coefs ... ", q);
#endif
  coefs = NewVector (q+1, "main");
  pColumns = new int[q+1];
  // read (index coefs) pairs
  for (int i=0; i<=q; ++i)
    scanf ("%d %lf\n", &pColumns[i], &coefs->ptr[i]);
  fprintf(stderr, "done\n");
  // read p and then n on next line
  scanf ("%d\n%d\n", &p, &n);
  // allocate buffer for input
  fprintf(stderr, "PredFilter: Reading %d obs for %d predictors\n", n, p);
  gBuffer = new float[p+1];

  // main calculation loop
  sse = 0.0; max = -1.0; min = 1.0;
  for (int line=0; line<n; ++line)
  { double error = 0.0;
    int y;
    LoadBuffer();
    prediction = VectorElement(coefs,0);
    for (int i=1; i < coefs->len; ++i)
      prediction += VectorElement(coefs, i) * BufferElement(pColumns[i]);
    error = BufferElement(0)-prediction;
    sse = sse + error * error;
    if (prediction > max) max = prediction;
    else if (prediction < min) min = prediction;
    y = (int) BufferElement(0);
    printf("%10.4f %3d\n", prediction, y);
  }
  // write a summary note to the screen in case piped
#ifdef WEIGHTS
  fprintf(stderr, "predict: Weighted Total SS = %10.3f with range [%10.3f, %10.3f]\n", sse, min, max);
#else
  fprintf(stderr, "predict: Total SS = %10.3f with range [%10.3f, %10.3f]\n", sse, min, max);
#endif
  return 1;
}

