/*
    Sweeper ... 
	This file is a part of sweeper, a program for identifying
	and estimating regression models, selecting from many possible 
	predictors without overfitting.

    Copyright (C) 1999,2002  Robert A. Stine

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License in the file COPYING for more details.
*/
// $Id: vector.c,v 1.17 2002/03/15 15:16:13 bob Exp $
/*
	15 Oct 98 ... Created 
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "vector.h"
#include "random.h"
#include "memory.h"

#define ExitOnNull(ptr,str,result) if(NULL==(ptr)){fprintf(stderr,"\nVECTOR: %s\n",(str));return (result);}
#define ExitOn(c, str, val) if (c){fprintf(stderr, "\nVECTOR: %s\n", (str)); return (val);}

#ifndef MIN
#define MIN(x,y) ((x) < (y)) ? (x) : (y)
#endif

#ifndef SQR
#define SQR(x) ((x) * (x))
#endif

#ifndef ABS
#define ABS(x) ((x)>0 ? (x) : (-(x)))
#endif

Vector NewVector (long n, char *creator)
{
  Vector pVec;

  pVec = (Vector) Allocate (sizeof(aVector), creator, "VECr");
  ExitOnNull(pVec, " could not allocate new vec.", NULL);
  pVec->ptr = (double *) Allocate (n * sizeof(double), creator, "VECp");
  if (pVec->ptr != NULL)
  {	pVec->len = n;
	return pVec;
  }
  else
  {	delete(pVec);
	fprintf(stderr, "VECTOR: could not allocate internal room for vector.\n");
	return NULL;
  }
}

// -----------------------  random vector ------------------------------------

static Vector
MakeRandomVector (long n, char *creator, double generator(void));

static Vector
MakeRandomVector (long n, char *creator, double generator(void))
{
  Vector vec;
  long i;

  if (HaveValidRandomState())
  { vec = NewVector(n, creator);
    if (vec)
      for (i=0; i<n; ++i)
	AssignVectorElement(vec, i, generator());
    return vec;
  } else
  {
    fprintf(stderr, "VECTOR: Invalid random state encountered.\n");
    return NULL;
  }
}	


Vector
NewUniformVector (long n, char *creator)
{
  return MakeRandomVector(n, creator, UniformRand);
}

Vector
NewNormalVector (long n, char *creator)
{
  return MakeRandomVector(n, creator, NormalRand);
}

/*
  Vector
  NewRandomVector (long n, char *creator)
  {
  Vector vec;
  long i;
  
  if (HaveValidRandomState())
  { vec = NewVector(n, creator);
  if (vec)
  for (i=0; i<n; ++i)
  AssignVectorElement(vec, i, UniformRand());
  return vec;
  } else
  {
  fprintf(stderr, "VECTOR: Invalid random state encountered.\n");
  return NULL;
  }
  }	
*/

// -----------------------------  files  -----------------------------

Vector 	RestoreVectorFromFile (FILE *file, char *creator)
{	Vector vec;
	long i, n;
	double x;
	
	fscanf (file, "%*s %ld\n", &n); // do not try to match
	if (n < 0)
	{	fprintf (stderr, "VECTOR: cannot read from file; length bad.\n");
		return NULL;
	} else
	{	vec = NewVector (n, creator);
		for (i=0; i<n; ++i)
		{	fscanf (file, "%lf ", &x);
			AssignVectorElement(vec, i, x);
		}
	}
	return (vec);
}


int	SaveVectorToFile (Vector v, FILE *file)
{	long i;

	fprintf (file, "v.len %ld\n", v->len);
	for(i=0; i<v->len; ++i)
		fprintf(file, "%f ", VectorElement(v,i));
	fprintf(file, "\n");
	return 0;
}


void	DeleteVector(Vector v)
{
  if (v)
  {
    if (v->ptr) Release(v->ptr);
    else fprintf(stderr, "VECTOR: attempt to delete unallocated vector.\n");
    v->len = 0;
    v->ptr = NULL;
    Release(v);
  } else
    fprintf(stderr,  "VECTOR: input vector is NULL.\n");
}

// ------------------------------------------------------	
	
long VectorLength (Vector v)
{
  if(NULL == v)
  {	fprintf(stderr, "VECTOR: pointer not allocated.\n");
	return 0;
  }
  return v->len;
}


#ifdef _CHECK_BOUNDS_
double VectorElement (Vector v, long index)
{
  if (NULL == v || NULL == v->ptr)
  {	fprintf(stderr, "VECTOR: pointer not allocated.\n");
	return 0.0;
  }
  if(index < 0 || index >= v->len)
  {	fprintf(stderr, "VECTOR: Index %ld out of range.\n", index);
	return 0.0;
  }
  return v->ptr[index];
}
#endif


int InitializeVector (Vector v, double initialValue)
{
  long i;

#ifdef _CHECK_BOUNDS
  if (NULL == v || NULL == v->ptr)
  {	fprintf(stderr, "VECTOR: pointer not allocated.\n");
	return 0;
  }
#endif
  for(i=0; i<VectorLength(v); ++i)
    AssignVectorElement(v,i,initialValue);
  return 1;
}


#ifdef CHECK_BOUNDS

int AssignVectorElement (Vector v, long index, double x)
{
  if (NULL == v || NULL == v->ptr)
  {	fprintf(stderr, "VECTOR: pointer not allocated.\n");
	return 0;
  }
  if (index < 0 || index >= v->len)
  {	fprintf(stderr, "VECTOR: Index %ld out of range in assignment.\n", index);
	return 0;
  } 
  v->ptr[index] = x;
  return 1;
}


int IncrementVectorElement (Vector v, long index, double x)
{
  if (index < 0 || index >= v->len)
  {	fprintf(stderr, "VECTOR: Index %ld out of range in increment.\n", index);
	return 0;
  }
  v->ptr[index] += x;
  return 1;
}

int DecrementVectorElement (Vector v, long index, double x)
{
  if (index < 0 || index >= v->len)
  {	fprintf(stderr, "VECTOR: Index %ld out of range in increment.\n", index);
	return 0;
  }
  v->ptr[index] -= x;
  return 1;
}

#endif

int DivideVectorElement (Vector v, long index, double divisor)
{
#ifdef _CHECK_BOUNDS
  if (index < 0 || index >= v->len)
  {	fprintf(stderr, "VECTOR: Index %ld out of range in division.\n", index);
	return 0;
  }
#endif
  v->ptr[index] /= divisor;
  return 1;
}

int MultiplyVectorElement	(Vector v, long index, double factor)
{
#ifdef _CHECK_BOUNDS
  if (index < 0 || index >= v->len)
  {	fprintf(stderr, "VECTOR: Index %ld out of range in multiply.\n", index);
	return 0;
  }
#endif
  v->ptr[index] *= factor;
  return 1;
}

int MultiplyVectors (Vector a, Vector b)
{
  long i;
  double *pb;
  
  if (a->len != b->len)
  {  fprintf(stderr, "VEC: Lengths of vectors to multiply do not match.\n");
     return 0;
  }
  pb = b->ptr;
  for (i=0; i<b->len; ++i)
    *pb++ = VectorElement(a,i) * VectorElement(b,i);
  return 1;
}

		
Vector CopyVector (Vector source, char *creator)
{
  Vector to;
  long i;
  
  ExitOnNull(to = NewVector(VectorLength(source), creator),
	     "unable to allocate copy", NULL);
  for (i=0; i<VectorLength(source); ++i)
    AssignVectorElement (to, i, VectorElement(source, i));
  return to;		
}

int
CopyVectorToDest (Vector src, Vector dest)
{
  long i;
  
  if (dest && (VectorLength(src)<=VectorLength(dest)))
    for (i=0; i<VectorLength(src); ++i)
      AssignVectorElement (dest, i, VectorElement(src, i));
  else
  { fprintf(stderr, "VEC: dest not allocated or insuffient room to copy.\n");
    return (0);
  }
  return 1;
}

int
CopyIndexedVector (IndexedVector *src, IndexedVector *dest)
{
  Vector sv, dv;

  dest->columnIndex = src->columnIndex;
  // check that dest is allocated
  sv = src->theVector;
  if (NULL == dest->theVector)
    dest->theVector = NewVector(VectorLength(sv), "CopyIndexedVector");
  if (dest->theVector)
  { dv = dest->theVector;
    CopyVectorToDest(sv,dv);
    return 1;
  }
  else
    return -1;
}



 
    
// ----------------------------------------------------

double VectorMean (Vector v)
{
  long i;
  double mean = 0.0;
  
  for (i=0; i<VectorLength(v); ++i)
    mean += VectorElement(v,i);
  mean /= (double) VectorLength(v);
  return (mean);
}

void CenterVector (Vector v, double center)
{
  long i;
  
  for (i=0; i<VectorLength(v); ++i)
    DecrementVectorElement(v, i, center);
}

double VectorSD (Vector v)
{
  long i;
  double mean, ss = 0.0;
  
  mean = VectorMean(v);
  for (i=0; i<VectorLength(v); ++i)
    ss += SQR(VectorElement(v,i) - mean);
  ss /= (double) (VectorLength(v) - 1);
  ss = sqrt(ss);
  return ss;
}

void ScaleVector (Vector v, double scale)
{
  double *pv = v->ptr;
  int     k  = v->len / 5;
  for (long i=0; i<k; ++i)
    {
      *pv = *pv / scale; ++pv;
      *pv = *pv / scale; ++pv;
      *pv = *pv / scale; ++pv;
      *pv = *pv / scale; ++pv;
      *pv = *pv / scale; ++pv;
    }
  for (int i=0; i<v->len % 5; ++i)
    {
      *pv = *pv / scale; ++pv;
    }
}

void StandardizeVector (Vector v)
{
  double mean, ss;
  long i;
  
  mean = VectorMean (v);
  for (i=0, ss=0.0; i < VectorLength(v); ++i)
  {	DecrementVectorElement(v, i, mean);
	ss += SQR(VectorElement(v,i));
  }
  ss /= (double) (VectorLength(v) - 1);
  ss = sqrt(ss);		// ss is now sd
  ScaleVector(v, ss);
}

void SquareVector (Vector v)
{
  long i;
  double x;

  for (i=0; i<VectorLength(v); ++i)
  { x = VectorElement(v,i);
    AssignVectorElement (v,i,x*x);
  }
}


double VectorDotProduct (Vector a, Vector b)
{
  long i;
  double dp=0.0;
#ifndef _CHECK_BOUNDS_
  double *aPtr, *bPtr;
  aPtr = a->ptr;
  bPtr = b->ptr;
#endif
  
  if (VectorLength(a) == VectorLength(b))
  {
    for (i=0; i<VectorLength(a); ++i)
    {
#ifdef _CHECK_BOUNDS_
      dp += VectorElement(a,i) * VectorElement(b,i);
#else
      dp += *aPtr++ * *bPtr++;
#endif
    }
  }
  return dp;
}


double	VectorCorrelation (Vector a, Vector b)
{	long i;
	double da, db, meanA, meanB, ssA, ssB, num, corr;
	
	ExitOn ((VectorLength(a)!=VectorLength(b)), "lengths do not match", 0.0);
	meanA = VectorMean(a);
	meanB = VectorMean(b);
	for(i=0, num=0.0, ssA=0.0, ssB=0.0; i<VectorLength(a); ++i)
	{	da = VectorElement(a,i)-meanA;
		db = VectorElement(b,i)-meanB;
		num += da * db;
		ssA += da * da;
		ssB += db * db;
	}
	corr = num / sqrt(ssA * ssB);
	return (corr);
}

double	VectorDistance(Vector a, Vector b, int norm)
{
  double diff, dist=0.0;
  long i,n;
  
  n = VectorLength(a);
  ExitOn ((n != VectorLength(b)), "lengths do not match", 0.0);
  switch (norm)
  {
  case 0:
    for(i=0; i <n; ++i)
    {	diff =VectorElement(a,i) - VectorElement(b,i);
	diff = ABS(diff);
	if (diff > dist) dist = diff;
    }
    break;
  case 1:
    for (i=0; i< n; ++i)
    {	diff =VectorElement(a,i) - VectorElement(b,i);
	diff = ABS(diff);
	dist += diff;
    }
    break;
  case 2:
    for (i=0; i< n; ++i)
    {	diff =VectorElement(a,i) - VectorElement(b,i);
	dist += SQR(diff);
    }
    break;
  default:	ExitOn(1, "improper norm requested", 0.0);
    break;
  }
  return (dist);
}


// ----------------------------------------------------


		
void PrintVector (Vector v, char *msg)
{
  PrintRawVector(v->len, v->ptr, msg);
}

void    PrintRawVector (long length, double *ptr, char *msg)
{
  long i;

  if (ptr)
  { printf ("\n --- %s [%ld] ---\n", msg, length);
    for (i=0; i<length; ++i)
      printf("[%3ld] %8.3f \n", i, ptr[i]);
  } else
    printf("\nVECTOR:  cannot print null vector.\n");
}


void CompactPrintVector(Vector v, char *msg)
{	
  printf (" --- %s[%ld] ---\n", msg, v->len);
  PrintVec(v);
}

void PrintRowVector (Vector v, long index)
{
  printf("[%4ld]", index);
  PrintVec(v);
}
	
void PrintVec(void *v)
{
  long i, min;
  Vector vec;
  
  if (v)
  {
    vec = (Vector) v;
    min = MIN(vec->len, 7);
    for (i = 0; i < min; ++i)
      printf(" %6.2f", VectorElement(vec, i));
    if (vec->len > 7)
      printf(" ... [%ld]", vec->len-7);
    printf("\n");
  } else
    printf ("VECTOR: cannot print null vector\n");
}





